"""Remote Demanders mechanisms

This module contains the different demanders mechanisms
that can be used to demand a TOKEN from a previously
discovered endpoint.


These are NOT grants, and are used by the remote grants
to discover the endpoint to connect to.

"""
